import logging
from typing import Dict
from http import HTTPStatus

import rasax.community.services.websocket_service as websocket_service
import rasax.community.utils.common as common_utils
from rasax.community.api.decorators import (
    rasa_x_scoped,
    validate_schema,
    inject_rasa_x_user,
)

from rasax.community.services.logs_service import LogsService
from rasax.community.services.workspaces_service import (
    WorkspacesDraftsCannotAccept,
    WorkspaceIsDefaultWorkspace,
    WorkspacesDraftsCannotReject,
    WorkspacesIdNotExists,
    WorkspacesNameExists,
    WorkspacesService,
)
from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse

logger = logging.getLogger(__name__)


def blueprint() -> Blueprint:
    """Create endpoints related to Workspaces management.

    Returns:
        Sanic `Blueprint` with endpoints.
    """
    workspaces_endpoints = Blueprint("workspaces_endpoints")

    @workspaces_endpoints.route("/workspaces", methods=["GET", "HEAD"])
    @rasa_x_scoped("workspaces.list", allow_api_token=True)
    async def get_workspaces(request: Request) -> HTTPResponse:
        """Return a list of all existing workspaces.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        workspaces = WorkspacesService.from_request(request).get_workspaces()
        return response.json(workspaces)

    @workspaces_endpoints.route("/workspaces", methods=["POST"])
    @rasa_x_scoped("workspaces.create", allow_api_token=True)
    @validate_schema("workspace")
    async def create_workspace(request: Request) -> HTTPResponse:
        """Create a new workspace with a specified name.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        try:
            workspace = WorkspacesService.from_request(request).create_workspace(
                request.json["name"]
            )

            return response.json(workspace, HTTPStatus.CREATED)
        except WorkspacesNameExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.CONFLICT,
                "WorkspaceCreationFailed",
                "Failed to create workspace.",
                details=e,
            )

    @workspaces_endpoints.route("/workspaces/drafts", methods=["PATCH"])
    @rasa_x_scoped("workspaces.drafts.update", allow_api_token=True)
    @validate_schema("workspace_drafts")
    async def patch_drafts_in_workspace(request: Request) -> HTTPResponse:
        """Patches drafts in a workspace.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        drafts_data = request.json.get("drafts")
        logs_service = LogsService.from_request(request)

        try:
            updated_drafts = logs_service.update_logs_in_workspace(drafts_data)
        except ValueError as e:
            return common_utils.error(
                HTTPStatus.NOT_FOUND,
                "DraftsUpdateFailed",
                "One of the drafts or workspaces doesn't exist",
                details=e,
            )

        if updated_drafts:
            websocket_service.send_message(
                websocket_service.Message(
                    websocket_service.MessageTopic.NLU,
                    name="workspaces_drafts_updated",
                    data={"drafts": updated_drafts},
                )
            )

        return response.json(updated_drafts)

    @workspaces_endpoints.route("/workspaces/<workspace_id>", methods=["PUT"])
    @rasa_x_scoped("workspaces.update", allow_api_token=True)
    @validate_schema("workspace")
    async def update_workspace(request: Request, workspace_id: int) -> HTTPResponse:
        """Updates the workspace name.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        try:
            workspace = WorkspacesService.from_request(request).update_workspace(
                workspace_id, request.json["name"]
            )

            return response.json(workspace, HTTPStatus.OK)
        except WorkspacesNameExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.CONFLICT,
                "WorkspaceUpdateFailed",
                "Failed to update workspace name.",
                details=e,
            )
        except WorkspacesIdNotExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.NOT_FOUND,
                "WorkspaceUpdateFailed",
                "Failed to update workspace name.",
                details=e,
            )
        except WorkspaceIsDefaultWorkspace as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.BAD_REQUEST,
                "WorkspaceUpdateFailed",
                "Failed to update workspace name.",
                details=e,
            )

    @workspaces_endpoints.route("/workspaces/<workspace_id>", methods=["DELETE"])
    @rasa_x_scoped("workspaces.delete", allow_api_token=True)
    async def delete_workspace(request: Request, workspace_id: int) -> HTTPResponse:
        """Deletes a workspace.

        Args:
            request: Incoming HTTP request.
            workspace_id: ID of the workspace to be deleted.

        Returns:
            HTTP response.
        """
        try:
            WorkspacesService.from_request(request).delete_workspace(
                workspace_id, delete_default_workspace=False
            )
            return response.empty()
        except WorkspacesIdNotExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.NOT_FOUND,
                "WorkspaceDeletionFailed",
                "Failed to delete workspace.",
                details=e,
            )
        except WorkspaceIsDefaultWorkspace as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.BAD_REQUEST,
                "WorkspaceDeletionFailed",
                "Failed to delete workspace.",
                details=e,
            )

    @workspaces_endpoints.route("/workspaces/drafts/rejected", methods=["POST"])
    @rasa_x_scoped("workspaces.drafts.reject", allow_api_token=True)
    @validate_schema("workspace_drafts_ids")
    async def reject_drafts_from_workspace(request: Request) -> HTTPResponse:
        """Rejects drafts from the workspace.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        ids = request.json["ids"]
        try:
            draft_ids = WorkspacesService.from_request(
                request
            ).reject_drafts_from_workspace(ids)

            if draft_ids:
                websocket_service.send_message(
                    websocket_service.Message(
                        websocket_service.MessageTopic.NLU,
                        name="workspaces_drafts_rejected",
                        data={"ids": draft_ids},
                    )
                )

            return response.json(draft_ids, HTTPStatus.OK)
        except WorkspacesDraftsCannotReject as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.BAD_REQUEST,
                "WorkspaceDraftsRejectFailed",
                "Failed to reject drafts.",
                details=e,
            )

    @workspaces_endpoints.route("/workspaces/drafts/accepted", methods=["POST"])
    @rasa_x_scoped("workspaces.drafts.accept", allow_api_token=True)
    @inject_rasa_x_user()
    @validate_schema("workspace_drafts_ids")
    async def accept_drafts_in_workspace(request: Request, user: Dict) -> HTTPResponse:
        """Accepts drafts in the workspace.

        This method will create new training data examples from the drafts
        specified in the paylod.

        Args:
            request: Incoming HTTP request.
            user: User that performs the action.

        Returns:
            HTTP response.
        """
        ids = request.json["ids"]
        try:
            draft_ids = WorkspacesService.from_request(
                request
            ).accept_drafts_in_workspace(ids, user)

            if draft_ids:
                websocket_service.send_message(
                    websocket_service.Message(
                        websocket_service.MessageTopic.NLU,
                        name="workspaces_drafts_accepted",
                        data={"ids": draft_ids},
                    )
                )

            return response.json(draft_ids, HTTPStatus.OK)
        except WorkspacesDraftsCannotAccept as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.BAD_REQUEST,
                "WorkspaceDraftsAcceptFailed",
                "Failed to accept drafts.",
                details=e,
            )

    return workspaces_endpoints
