import json
from datetime import datetime
from typing import Text, List, Dict, Any, Optional, Union

import rasax.community.utils.common as common_utils
from rasax.community.database.assistant import (
    BusinessLogicFlow,
    BusinessLogicFlowElement,
    SampleConversation,
    SampleConversationItem,
)
from rasax.community.database.service import DbService
from sanic.request import Request
from sqlalchemy.exc import IntegrityError


class AssistantService(DbService):
    """A service used to work with assistant-related models."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "AssistantService":
        """Construct `InsightService` from an incoming HTTP request.

        Args:
            request: Incoming HTTP request.
            **kwargs: Other key-value args, not used.

        Returns:
            Constructed `InsightService` object.
        """
        return AssistantService(request.ctx.db_session)

    def fetch_flows(self, limit: int = 30, offset: int = 0) -> common_utils.QueryResult:
        """Fetches Business Logic Flows.

        Args:
            limit:  limit to use for the query, defaults to 30
            offset: offset to use for the query, defaults to 0

        Returns:
            QueryResult with `BusinessLogicFlow`s matching the query.
        """
        query = self.query(BusinessLogicFlow).order_by(
            BusinessLogicFlow.updated_at.desc()
        )
        count = query.count()
        query = query.limit(limit).offset(offset)

        result = [flow.as_dict() for flow in query.all()]
        return common_utils.QueryResult(result, count)

    def get_flow(self, flow_id: Any) -> Optional[BusinessLogicFlow]:
        """Gets Business Logic Flow by id.

        Args:
            flow_id: ID of the Business Logic Flow.

        Return:
            BusinessLogicFlow or `None` if none could be found with the given ID.
        """
        try:
            lookup_id = int(flow_id)
            return self.query(BusinessLogicFlow).get(lookup_id)
        except (ValueError, TypeError):
            return None

    def create_flow(
        self,
        name: Text,
        elements: List[Dict[Text, Any]],
        description: Text = None,
        created_at: datetime = None,
        updated_at: datetime = None,
    ) -> BusinessLogicFlow:
        """Creates a new Business Logic Flow.

        Args:
            name:        a unique name.
            elements:    a list of flowchart steps and arrows as dicts.
            description: optional value for further describing the business logic flow.
            created_at:  optional value when it was created, if missing uses `datetime.utcnow()`
            updated_at:  optional value when it was updated, if missing uses `datetime.utcnow()`

        Raises:
            DuplicateBusinessLogicFlowException: When a business logic flow with the same name already exists.

        Returns:
            Created `BusinessLogicFlow`.
        """
        try:
            flow = BusinessLogicFlow(
                name=name,
                description=description,
                created_at=created_at,
                updated_at=updated_at,
            )
            flow.elements = self._parse_flow_elements(elements)
            self.add(flow)

            self.flush()
            return flow
        except IntegrityError:
            self.rollback()
            raise DuplicateBusinessLogicFlowException()

    def update_flow(
        self,
        flow_id: Any,
        name: Text,
        elements: List[Dict[Text, Any]],
        description: Text,
    ) -> BusinessLogicFlow:
        """Updates a Business Logic Flow.

        Args:
            flow_id:     ID of the Business Logic Flow.
            name:        a unique name.
            description: a string for further describing the business logic flow.
            elements:    a list of flowchart steps and arrows as dicts.

        Raises:
            ValueError: when a flow with the given ID cannot be found.
            BusinessLogicFlowParseException: when the given elements cannot be parsed.

        Returns:
            Updated `BusinessLogicFlow`.
        """
        flow = self.get_flow(flow_id)

        if flow is None:
            raise ValueError()

        try:
            flow.name = name
            flow.description = description
            self._merge_elements(flow, elements)

            self.flush()
            self.session.refresh(flow)
            return flow
        except IntegrityError:
            self.session.rollback()
            raise DuplicateBusinessLogicFlowException()

    def delete_flow(self, flow_id: Any) -> None:
        """Deletes a Business Logic Flow if it exists.

        Args:
            flow_id: ID of the Business Logic Flow.

        Raises:
            ValueError: when a flow with the given ID cannot be found.
        """
        flow = self.get_flow(flow_id)

        if flow is None:
            raise ValueError()

        self.delete(flow)
        self.flush()

    def fetch_sample_conversations(
        self, limit: int = 30, offset: int = 0
    ) -> common_utils.QueryResult:
        """Fetches Sample Conversations.

        Args:
            limit:  limit to use for the query, defaults to 30
            offset: offset to use for the query, defaults to 0

        Returns:
            QueryResult with `SampleConversation`s matching the query.
        """
        query = self.query(SampleConversation).order_by(
            SampleConversation.updated_at.desc()
        )
        count = query.count()
        query = query.limit(limit).offset(offset)

        result = [conv.as_dict() for conv in query.all()]
        return common_utils.QueryResult(result, count)

    def get_sample_conversation(
        self, sample_conversation_id: Any
    ) -> Optional[SampleConversation]:
        """Gets Sample Conversation by id.

        Args:
            sample_conversation_id: ID of the Sample Conversation.

        Return:
            SampleConversation or `None` if none could be found with the given ID.
        """
        try:
            lookup_id = int(sample_conversation_id)
            return self.query(SampleConversation).get(lookup_id)
        except (ValueError, TypeError):
            return None

    def create_sample_conversation(
        self,
        name: Text,
        description: Text = None,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
    ) -> SampleConversation:
        """Creates a sample conversation.

        Args:
            name:        The name of the conversation.
            description: optional value for further describing the sample conversation.
            created_at:  optional value when it was created, if missing uses `datetime.utcnow()`
            updated_at:  optional value when it was updated, if missing uses `datetime.utcnow()`

        Returns:
            Created `SampleConversation`
        """
        try:
            sample_conversation = SampleConversation(
                name=name,
                description=description,
                created_at=created_at,
                updated_at=updated_at,
            )
            self.session.add(sample_conversation)
            self.session.flush()
            return sample_conversation
        except IntegrityError:
            self.session.rollback()
            raise DuplicateSampleConversationException()

    def update_sample_conversation(
        self, sample_conversation_id: Any, name: Text, description: Text,
    ) -> SampleConversation:
        """Updates a Sample Conversation.

        Args:
            sample_conversation_id: ID of the Business Logic Flow.
            name:                   a unique name.
            description:            a string for further describing the business logic flow.

        Raises:
            ValueError: when a sample conversation with the given ID cannot be found.

        Returns:
            Updated `SampleConversation`.
        """
        conversation = self.get_sample_conversation(sample_conversation_id)

        if conversation is None:
            raise ValueError()

        try:
            conversation.name = name
            conversation.description = description

            self.flush()
            return conversation
        except IntegrityError:
            self.session.rollback()
            raise DuplicateSampleConversationException()

    def create_sample_conversation_message(
        self, type_: Text, text: Text, sample_conversation_id: Any
    ) -> SampleConversationItem:
        """Creates message to an existing sample conversation.

        Args:
            type_: The type of the message (`user_message` or `bot_message`).
            text: The text of the message.
            sample_conversation_id: The ID of the sample conversation this message is part of.

        Returns:
            The created `SampleConversationItem`.

        Raises:
            `ValueError` when one or more given values are invalid or missing.
        """
        conversation = self.get_sample_conversation(sample_conversation_id)
        if conversation is None:
            raise ValueError("sample conversation does not exist")

        message = SampleConversationItem(
            type=type_, text=text, conversation=conversation
        )
        self.session.add(message)
        self.session.flush()
        conversation.updated_at = message.updated_at

        return message

    def delete_sample_conversation(self, sample_conversation_id: Any) -> None:
        """Deletes a Sample Conversation if it exists.

        Args:
            sample_conversation_id: ID of the Sample Conversation.

        Raises:
            ValueError: when a sample conversation with the given ID cannot be found.
        """
        conversation = self.get_sample_conversation(sample_conversation_id)

        if conversation is None:
            raise ValueError()

        self.delete(conversation)
        self.flush()

    def delete_sample_conversation_message(
        self, sample_conversation_item_id: Any
    ) -> None:
        """Deletes a Sample Conversation Message if it exists.

        Args:
            sample_conversation_item_id: ID of the Sample Conversation Item.

        Raises:
            ValueError: when a sample conversation item with the given ID cannot be found.
        """
        try:
            lookup_id = int(sample_conversation_item_id)
            item = self.query(SampleConversationItem).get(lookup_id)

            if item is None:
                raise ValueError()

            self.delete(item)
            self.flush()
        except TypeError:
            raise ValueError()

    @staticmethod
    def _parse_flow_elements(
        elements: List[Dict[Text, Any]]
    ) -> List[BusinessLogicFlowElement]:
        """Parses flow elements and returns initialized `BusinessLogicFlowElement` objects."""
        flow_elements = []

        for element in elements:
            flow_element = BusinessLogicFlowElement()
            try:
                flow_element.update_with_data(element)
            except ValueError as e:
                raise BusinessLogicFlowParseException(str(e))
            flow_elements.append(flow_element)

        return flow_elements

    def _merge_elements(
        self, flow: BusinessLogicFlow, elements: List[Dict[Text, Any]]
    ) -> None:
        """Parses flow elements and merges them with elements of an existing business logic flow."""
        parsed_elements = self._parse_flow_elements(elements)

        def _find_by_element_id(
            flow_elements: List[BusinessLogicFlowElement], element_id: Text
        ) -> Union[None, BusinessLogicFlowElement]:
            for element in flow_elements:
                if element.element_id == element_id:
                    return element
            return None

        existing_element_ids = {e.element_id for e in flow.elements}
        input_element_ids = {e.element_id for e in parsed_elements}
        element_ids_to_delete = existing_element_ids - input_element_ids
        for delete_id in element_ids_to_delete:
            existing_element = _find_by_element_id(flow.elements, delete_id)
            self.delete(existing_element)

        for input_element in parsed_elements:
            existing_element = _find_by_element_id(
                flow.elements, input_element.element_id
            )
            if existing_element is None:
                flow.elements.append(input_element)
            else:
                existing_element.update_with_data(json.loads(input_element.data))


class DuplicateBusinessLogicFlowException(Exception):
    """Exception for a business logic flow with a duplicate name."""

    pass


class BusinessLogicFlowParseException(Exception):
    """Exception when parsing invalid business logic flow elements."""

    pass


class DuplicateSampleConversationException(Exception):
    """Exception for a sample conversation with a duplicate name."""

    pass
